from decimal import Decimal

def thous(x:float)-> str:
    return('{0:n}'.format(Decimal(x)))

def cleanCurlyAround(s:str)->str:
    """Removes curly braces"""
    if s and (s[0]=='{') and (s[-1]=='}'):
        return(s[1:-1])
    return(s)




from tqdm import tqdm
from pybtex.database.input import bibtex
from pybtex.database import Person
import re
from titlecase import titlecase
from refcliq.bibtex import parse
from refcliq.util import thous, cleanCurlyAround

# _citePattern=re.compile(r"{?(?P<author>[\w\s\.\(\)-]*?)]?(, (?P<year>\d{4}))?, (?P<journal>.*?)(, (?P<vol>V(, )?[^,]+))?(, (?P<page>P(, )?[^,]+))?(, [DOI ^,]+(?P<doi>10.( )*\d{4,9}/[ ,._;()/<>:#\\A-Z0-9\-\+]*))?((\. )|(\.})|(\.\Z)|(}\Z))", flags=re.IGNORECASE)
_citePattern=re.compile(r"{?(?P<author>[\w\s\.\(\)-]*?)?(, )?(?P<year>\d{4})?(, (?P<journal>.*?))(, (?P<vol>V[, ]*[^,]+))?(, (?P<page>P[, ]*[^,]+))?(, [DOIRG/ ^,]+(?P<doi>.*))?((\. )|(\.})|(\.\Z)|(}\Z))", flags=re.IGNORECASE)
_listPattern=re.compile(r'\{\[\}(.*?)(,.*?)+\]')

def _properName(name:str)->str:
    """
    Properly formats the reference name. While it understands van/der, it breaks
    for institution names, because we can't tell them apart from people's names.
    """        
    vals=name.split(' ')
    lasts=[vals[0].lower(),]
    i=1
    while (lasts[-1].lower() in ['de','der','von','van']):
        lasts.append(vals[i].lower())
        i+=1
    lasts[-1]=titlecase(lasts[-1])
    last=' '.join([w for w in lasts])
    rest=[]
    for v in vals[i:]:
        if all([c.isupper() for c in v]): #Initials - JE
            rest.extend([c for c in v])
        else:
            rest.append(titlecase(v.lower()))
    return(last+", "+' '.join(rest).replace(".",""))

def dict_update_notNone_inplace(D:dict,k:any,val:any):
    """
        Adds the pair (k,val) to the dictionary D if val is not None.
        This helps keep the resulting json smaller (by not adding a bunch of null keys).
    """
    if val is not None:
        D[k]=val

def split_reference(reference:str)->dict:
    """
    Generates a dictionary with the info present on the _single_ reference line.

    references: raw text from "cited-references" WoS's .bib (with \n!)
    return: {author, year ,journal, vol, page, doi}, if they are not null.
    """
    #removes the \_ from DOIs
    ref=reference.replace(r"\_","_")
    #removes the non-list {[} ]
    ref=re.sub(r"\{\[\}([^,\]]*?)\]",r"\1",ref)
    #replaces inner lists {[} X, Y] with X
    ref=_listPattern.sub(r'\2',ref)  #the first part is usually the same but in chinese/etc
    
    match=_citePattern.search(ref)
    if match:
        doi = match.group('doi')
        if doi:#removes whitespaces in the DOI, yes, we have them
            doi=doi.translate(str.maketrans('','',' '))

        article={'authors' : [Person(string=_properName(match.group('author'))),]}
        dict_update_notNone_inplace(article, 'year', match.group('year'))
        dict_update_notNone_inplace(article, 'journal',titlecase(match.group('journal')))
        dict_update_notNone_inplace(article, 'vol', match.group('vol'))
        dict_update_notNone_inplace(article, 'page', match.group('page'))
        dict_update_notNone_inplace(article, 'doi', doi)
    # we know this is a reference. It might be only the name of the publication
    else:
        article={'authors':[],
                 'journal': titlecase(reference),
                 }
    return(article)

def extract_article_info(fields, people, references:list)->dict:
    """
    Creates a dict with the information from the bibtex fields.
    "references" is the raw Cited-References field from WoS' with \n s
    """

    abstract = cleanCurlyAround(fields.get('abstract',''))
    if ' (C) ' in abstract:
        abstract = abstract.split(' (C) ')[0]

    refs=[]
    for r in references:
        if ('in press' in r.lower()):
            better_ref=re.sub('in press','',r,flags=re.IGNORECASE)
            refs.append(split_reference(better_ref))
            refs[-1]['inPress']=True
        else:
            refs.append(split_reference(r))

    doi=fields.get('doi', None)
    if doi:
        doi=cleanCurlyAround(doi.lower())
        
    affiliation = fields.get('Affiliation',None)
    if affiliation:
        affiliation=cleanCurlyAround(affiliation).replace(r'\&','&')
    
    ret={'authors': people.get("author",[])}
    dict_update_notNone_inplace(ret, 'Affiliation', affiliation)
    dict_update_notNone_inplace(ret, 'year', cleanCurlyAround(fields.get('year', None)))
    dict_update_notNone_inplace(ret, 'doi', doi)
    dict_update_notNone_inplace(ret, 'title', titlecase(cleanCurlyAround(fields.get("title", None))))
    dict_update_notNone_inplace(ret, 'journal', titlecase(cleanCurlyAround(fields.get('series', fields.get('journal', None) ))))
    dict_update_notNone_inplace(ret, 'volume', cleanCurlyAround(fields.get('volume', None)))
    dict_update_notNone_inplace(ret, 'pages', cleanCurlyAround(fields.get('pages', None)))
    dict_update_notNone_inplace(ret, 'references', refs)
    dict_update_notNone_inplace(ret, 'number', cleanCurlyAround(fields.get('number', None)))
    dict_update_notNone_inplace(ret, 'abstract', abstract )
    return(ret)

def import_bibs(filelist:list) -> list:
    """
    Takes a list of bibtex files and returns entries as a list of dictionaries
    representing the info on each work
    """
    articles = []
    references_field="Cited-References"
    print('Reading .bibs')
    for filename in tqdm(filelist):
        try:
            #since pybtex removes the \n from this field, we do it ourselves
            #(but we are not fully replacing pybtex because of the extra consistency checks it has)
            references=parse(filename,keepOnly=[references_field,])
            for k in references:
                if (references_field not in references[k]):
                    references[k][references_field]=[]
                else:
                    references[k][references_field]=[x.strip() for x in references[k][references_field].split('\n')]

            bibdata = {}
            parser = bibtex.Parser()
            bibdata = parser.parse_file(filename)

            for bib_id in bibdata.entries:
                articles.append(extract_article_info(bibdata.entries[bib_id].fields,
                                                bibdata.entries[bib_id].persons,
                                                references[bib_id][references_field]))

        except:
            print('Error with the file ', filename)
            raise

    print('Imported {0} articles.'.format(thous(len(articles))))
    return(articles)



#!/usr/bin/env python
# encoding: utf-8
"""
RefCliq is a (full) rewrite of Neal Caren's original script.

git: https://github.com/fabioasdias/RefCliq

The idea is the same, but with improved article matching, a more stable
clustering method (the same python-louvain community, but considering the number
of co-citations on the edges), geo-coding support for the authors, and a web
interface for the visualization of the results.


Original: https://github.com/nealcaren/RefCliq 
Created by Neal Caren on June 26, 2013. 
neal.caren@gmail.com
"""

from community import best_partition
from argparse import ArgumentParser

from refcliq.citations import CitationNetwork
from refcliq.util import thous

import json
from tqdm import tqdm
import networkx as nx
from networkx.readwrite import json_graph
from os.path import exists
import pickle

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument("-o", type=str,
                        help="Output file to save, defaults to 'clusters.json'.",
                        dest="output_file", default='clusters.json')
    parser.add_argument("--cites", type=int,
                        help="Minimum number of citations for an article to be included, defaults to 2.",
                        dest="cites", default=2)
    parser.add_argument("-k", type=str,
                        help="Google maps API key. Necessary for precise geocoding.",
                        dest="google_key", default='')
    parser.add_argument("--graphs",  action="store_true",
                        help="Saves graph drawing information for the cluster.",
                        dest="graphs", default=False)
    parser.add_argument("files", nargs='+',
                        help="List of .bib files to process")

    options = parser.parse_args()

    citation_network = CitationNetwork()
    citation_network.build(
        options.files, google_key=options.google_key, min_citations=options.cites)

    citation_network.compute_keywords()

    print(thous(len(citation_network))+' different references with ' +
          thous(len(citation_network.edges()))+' citations.')

    co_citation_network = citation_network.cocitation(
        count_label="count", copy_data=False, min_cocitations=options.cites)

    for n in citation_network:
        citation_network.node[n]['data']['original_cc'] = -1

    for i, gg in enumerate(nx.connected_components(co_citation_network)):
        for n in gg:
            citation_network.node[n]['data']['original_cc'] = i

    print('\nPartitioning (no progress bar for this - sorry!)\n')
    partition = best_partition(
        co_citation_network, weight='count', random_state=7)  # deterministic

    output = {'geocoded': options.google_key != ''}

    parts = {}
    for n in partition:
        if partition[n] not in parts:
            parts[partition[n]] = []
        parts[partition[n]].append(n)

    graphs = {}

    print('Per cluster analysis/data (centrality)')
    for p in tqdm(parts):
        subgraph = co_citation_network.subgraph(parts[p])
        centrality = nx.degree_centrality(subgraph)
        if options.graphs:
            topo = nx.Graph()
            topo.add_nodes_from(subgraph)
            topo.add_weighted_edges_from(subgraph.edges(data='count'))
            graphs[p] = json_graph.node_link_data(topo)

        for n in centrality:
            citation_network.node[n]['data']['centrality'] = centrality[n]

    output['partitions'] = parts
    output['graphs'] = graphs

    print('Saving')
    articles = {}
    for n in tqdm(citation_network.nodes()):
        # if the work isn't cited or doesn't cite anything useful (aka is never going to show up in the interface)
        if (n not in co_citation_network) and (all([p not in co_citation_network for p in citation_network.successors(n)])):
            continue
        articles[n] = citation_network.node[n]['data']
        articles[n]['cites_this'] = [
            p for p in citation_network.predecessors(n)]
        articles[n]['cited_count'] = len(articles[n]['cites_this'])
        articles[n]['references'] = [p for p in citation_network.successors(n)]
        articles[n]['reference_count'] = len(articles[n]['references'])
        articles[n]['authors'] = [
            {'last': x.last_names, 'first': x.first_names} for x in articles[n]['authors']]

    output['articles'] = articles

    outName = options.output_file
    if not outName.endswith('.json'):
        outName = outName+'.json'

    with open(outName, 'w') as fout:
        json.dump(output, fout)  # , indent=4, sort_keys=True)

    print('Run "rc_vis.py {0}" to view the results.'.format(outName))



#!/usr/bin/env python
# encoding: utf-8

import http.server
import socketserver
import sys
import webbrowser
from tempfile import TemporaryDirectory
from shutil import copytree, copy
from os import chdir
from os.path import join
from imp import find_module


template = 'template'

if __name__ == "__main__":

    if len(sys.argv) != 2:
        print('\nUsage:\n {0} result.json\n Where "result.json" is the result to be visualized.'.format(
            sys.argv[0]))
        exit(-1)

    with TemporaryDirectory() as basefolder:
        print(find_module('refcliq'))
        newRoot=join(basefolder,'refcliq')
        copytree(join(find_module('refcliq')[1], template), newRoot)
        copy(sys.argv[1], join(newRoot, 'data.json'))
        chdir(newRoot)

        PORT = 8080
        Handler = http.server.SimpleHTTPRequestHandler
        url = "http://localhost:{0}/".format(PORT)
        webbrowser.open_new_tab(url)
        with socketserver.TCPServer(("", PORT), Handler) as httpd:
            print('Serving at {0}\n\nPress Ctrl+C (or Command+dot) to stop.\n\n')
            httpd.serve_forever()

        



import re
from refcliq.util import cleanCurlyAround

_firstLinePattern=re.compile(r"@(?P<kind>.*?){([\s]*(?P<id>[^,\s]+),)")
_fieldPattern=re.compile(r"(?P<name>[\w-]+?)[\s]*=[\s]*({(?P<content>.*)})",    flags=re.IGNORECASE|re.DOTALL) 
def parse(bibfile:str, keepOnly:list=None)->dict:
    """
    bibfile: path to a .bib file.
    
    Returns: a dictionary where the keys are the entry IDs and the content are
    the entries, using the same fields.

    If keepOnly is not None, only returns those fields.
    """
    ret={}
    with open(bibfile,'r') as fin:
        currentEntry=None
        for line in fin:
            match=_firstLinePattern.search(line)    
            if (match):
                currentEntry=match.group("id")
                ret[currentEntry]={}
                currentField=''
                openBraces=0
                continue
            #blank spaces/lines, preambles, etc
            if (currentEntry is None): 
                continue
            openBraces+=(line.count('{')-line.count('}'))
            currentField=currentField+line
            if openBraces > 0:
                continue
            else:
                match=_fieldPattern.search(currentField)
                if (match):
                    name=match.group("name")
                    if (name is not None) and ((keepOnly is None) or (name in keepOnly)):
                        content=cleanCurlyAround(match.group("content")).strip()
                        ret[currentEntry][match.group("name")]=content
                currentField=''
            
    return(ret)




import networkx as nx
from networkx.readwrite import json_graph
from fuzzywuzzy.fuzz import ratio
from tqdm import tqdm
from itertools import combinations
from collections import Counter
from math import floor
from string import punctuation
from collections import Counter
from nltk.corpus import stopwords
from nltk import WordNetLemmatizer, download, bigrams
from math import log
from nltk.tokenize import word_tokenize
from nltk.stem import snowball

from refcliq.geocoding import ArticleGeoCoder
from refcliq.preprocess import import_bibs

from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
from scipy.sparse import coo_matrix

from os.path import exists
import klepto
import pickle

download('stopwords')
download('punkt')
download('wordnet')

stemmer = snowball.EnglishStemmer()
remove_punct = str.maketrans('', '', punctuation)


def tokens_from_sentence(sentence: str, remove_duplicates: bool=True)->list:
    """
      Returns a list of "important" words in a sentence.
      Only works in English but returned tokens may not be proper English.
    """
    stop_words = stopwords.words('english')
    words = [word.translate(remove_punct).lower()
             for word in word_tokenize(sentence)]
    words = [stemmer.stem(word) for word in words if (
        word not in stop_words) and (word != '')]
    if remove_duplicates:
        return(list(set(words)))
    else:
        return(words)


# https://medium.com/analytics-vidhya/automated-keyword-extraction-from-articles-using-nlp-bfd864f41b34
def sort_coo(coo_matrix):
    tuples = zip(coo_matrix.col, coo_matrix.data)
    return sorted(tuples, key=lambda x: (x[1], x[0]), reverse=True)


def extract_topn_from_vector(feature_names, sorted_items, topn=10):
    """get the feature names and tf-idf score of top n items"""

    score_vals = []
    feature_vals = []

    # word index and corresponding tf-idf score

    to_remove = set()
    for idx, score in sorted_items:
        # do not add keywords that are contained in other keywords: sao paulo, sao, paulo.
        skip = False
        # we already added that

        for i, keywords in enumerate(feature_vals):
            for already in keywords.split():
                for toAdd in feature_names[idx].split():
                    if toAdd in already:
                        skip = True
                        # let's keep the longest keyword
                        if len(feature_vals[i]) < len(feature_names[idx]):
                            feature_vals[i] = feature_names[idx]

        if not skip and(feature_names[idx] not in feature_vals):
            score_vals.append(round(score, 6))
            feature_vals.append(feature_names[idx])

            for i in range(len(feature_vals)):
                for j in range(i+1, len(feature_vals)):
                    if feature_vals[i] == feature_vals[j]:
                        score_vals[i] += score_vals[j]
                        to_remove.add(j)
            if (len(score_vals)-len(to_remove)) == topn:
                break

    for j in sorted(to_remove, reverse=True):
        del(feature_vals[j])
        del(score_vals[j])

    # create a tuples of feature,score
    results = zip(feature_vals, score_vals)

    return(list(results))


def _merge_keywords(keywords: list, how_many_works: int)->list:
    """
        Removes duplicate keywords from a list, updating the tf-idf (summing the
        occurences). How_many_works is used to re-normalize the values.
    """
    merged = {}
    for k, v in keywords:
        if k not in merged:
            merged[k] = v
        else:
            merged[k] += v
    return([(k, merged[k]/how_many_works) for k in merged])

# @profile


def same_article(a1: dict, a2: dict)->bool:
    """
    Tests if two articles are the same based on their info. 

    Uses fuzzy string matching and the Person structure from pybtex.
    """

    # usefulFields=list(set([k for k in a1 if a1[k]]).intersection(set([k for k in a2 if a2[k]])))

    if ('year' in a1) and ('year' in a2):
        if (a1['year'] != a2['year']):
            return(False)

    # two articles can have the same doi!
    # if 'doi' in usefulFields:
    #     if (a1['doi']!=a2['doi']):
    #         return(False)

    # article from references only have one author that we know of
    L1 = len(a1['authors'])
    L2 = len(a2['authors'])
    if (L1 != 1) and (L2 != 1) and (L1 != L2):
        return(False)

    for field in ['title', 'journal', 'page', 'vol']:
        if (field in a1) and (field in a2) and (ratio(a1[field], a2[field]) <= 80):
            return(False)

    for i in range(min([L1, L2])):
        if (ratio(str(a1['authors'][i]), str(a2['authors'][i])) <= 80) and (ratio(' '.join(a1['authors'][i].last_names), ' '.join(a2['authors'][i].last_names)) <= 80):
            return(False)

    return(True)


class CitationNetwork(nx.DiGraph):
    def __init__(self):
        nx.DiGraph.__init__(self)
        # self._G=nx.DiGraph() #the network
        self._year = {'None': set()}  # indexes
        self._authors = {1: set()}
        self._journal = {0: set()}
        self._title = {0: set()}
        self._authorName = {0: set()}  # None might be a part of a name
        self._equivalentDOIs = {}  # yes, one paper can have more than one DOI

    def save(self, filename: str):
        """Saves the citation network structure to filename"""
        graph = json_graph.node_link_data(self)
        d = klepto.archives.hdf_archive(filename, {'year': self._year,
                                                   'authors': self._authors,
                                                   'title': self._title,
                                                   'authorName': self._authorName,
                                                   'DOIs': self._equivalentDOIs,
                                                   'graph': graph}, cached=False, serialized=True)
        d.dump()

    def load(self, filename: str):
        """Loads the citation network structure from filename"""
        d = klepto.archives.hdf_archive(
            filename, cached=False, serialized=True)

        d.load('graph')
        G = json_graph.node_link_graph(d['graph'])
        self.add_nodes_from(G.nodes(data=True))
        self.add_edges_from(G.edges(data=True))

        for saved, internal in [('DOIs', '_equivalentDOIs'), ('authorName', '_authorName'), ('title', '_title'), ('authors', '_authors'), ('year', '_year')]:
            d.load(saved)
            self.__dict__[internal] = d[saved]

    # @profile
    def build(self, bibs: list, google_key: str='', min_citations: int=2):
        """
        Builds a directed graph to represent the citation network of the file list bibs.
        Ignores any references cited less than min_citation times, along with the citing papers.
        """

        geoCoder = ArticleGeoCoder(google_key)
        articles = import_bibs(bibs)

        print('citation network - Full citation in the .bibs')
        for article in tqdm(articles):
            citing = self.find(article)

        print('citation network - Cited-References')
        for article in tqdm(articles):
            citing = self.find(article)
            for cited_article in article['references']:
                cited = self.find(cited_article)
                self.add_edge(citing, cited)

        to_remove = []
        for n in self:
            # if this article has fewer than min_citations citations AND doesn't cite anything that has more than min_citations
            if (len(list(self.predecessors(n))) < min_citations) and ((len(list(self.successors(n))) == 0) or (max([len(list(self.predecessors(x))) for x in self.successors(n)]) < min_citations)):
                to_remove.append(n)

        for n in to_remove:
            self.remove(n)

        geoCoder.add_authors_location_inplace(self)
        print('Outgoing geocoding calls ', geoCoder._outgoing_calls)

    def remove(self, n: any):
        """
            Removes the node n from the network and the indices.
        """
        self._year[self.node[n]['index']['year']].remove(n)
        self._authors[self.node[n]['index']['authors']].remove(n)
        for index in self.node[n]['index']['journal']:
            self._journal[index].remove(n)
        for index in self.node[n]['index']['title']:
            self._title[index].remove(n)
        for index in self.node[n]['index']['name']:
            self._authorName[index].remove(n)
        self.remove_node(n)

    def add(self, article: dict, replaceNode: str=None)->str:
        """
        Adds a new citation to the network.
        if replaceNode is given, replaces that node in the network
        (this is used to update the name to use a previously unknown DOI)

        Returns the node.
        """

        if ('doi' in article) and (article['doi'] is not None):
            ID = article['doi']
            if replaceNode:
                # print('Replacing {0} with {1}'.format(replaceNode,ID))
                n = replaceNode
                if n[0] != '-':  # aka this is already a DOI!
                    self._equivalentDOIs[ID] = n  # mark as equivalent
                    return(n)  # short-circuits the rest

                for nn in self.predecessors(n):
                    self.add_edge(nn, ID)
                for nn in self.successors(n):
                    self.add_edge(ID, nn)
                # removes from the indexes

                for field in [f for f in self.node[n]['data'] if self.node[n]['data'][f]]:
                    if (field == 'abstract') and (self.node[n]['data'][field] != '') and ((field not in article) or (article[field] == '')):
                        article[field] = self.node[n]['data'][field]
                    elif (field == 'authors') and len(self.node[n]['data']['authors']) > len(article['authors']):
                        article['authors'] = self.node[n]['data']['authors'][:]
                    elif (field not in article) or (not article[field]):
                        article[field] = self.node[n]['data'][field]

                self.remove(n)
        else:
            ID = '-'+str(len(self.nodes()))  # flags as non-DOI

        self.add_node(ID)
        self.node[ID]['data'] = {**article}

        # store which index in the node to make update easier
        self.node[ID]['index'] = {}

        # if ('doi' not in article) or (article['doi'] is None):
        #     self._noDOI.append(ID)

        if 'year' in article:
            yearIndex = article['year']
            self.node[ID]['index']['year'] = yearIndex
            if yearIndex not in self._year:
                self._year[yearIndex] = set()
            self._year[yearIndex].add(ID)
        else:
            self.node[ID]['index']['year'] = 'None'
            self._year['None'].add(ID)

        # authors is the only field that always exists.
        authorIndex = len(article['authors'])
        self.node[ID]['index']['authors'] = authorIndex
        if authorIndex not in self._authors:
            self._authors[authorIndex] = set()
        self._authors[authorIndex].add(ID)

        if ('journal' in article) and (article['journal'] is not None):
            tokens = tokens_from_sentence(article['journal'])
            self.node[ID]['index']['journal'] = tokens
            for token in tokens:
                if token not in self._journal:
                    self._journal[token] = set()
                self._journal[token].add(ID)
        else:
            self._journal[0].add(ID)
            self.node[ID]['index']['journal'] = [0, ]

        if ('title' in article) and (article['title'] is not None):
            tokens = tokens_from_sentence(article['title'])
            self.node[ID]['index']['title'] = tokens
            for token in tokens:
                if token not in self._title:
                    self._title[token] = set()
                self._title[token].add(ID)
        else:
            self._title[0].add(ID)
            self.node[ID]['index']['title'] = [0, ]

        if ('authors' in article) and len(article['authors']) > 0:
            self.node[ID]['index']['name'] = set()
            for author in article['authors']:
                for name in author.last_names:
                    token = name.lower()
                    self.node[ID]['index']['name'].add(token)
                    if token not in self._authorName:
                        self._authorName[token] = set()
                    self._authorName[token].add(ID)
        else:  # no authors
            self._authorName[0].add(ID)
            self.node[ID]['index']['name'] = [0, ]
        return(ID)
    # @profile

    def _find_article_no_doi(self, article: dict):
        """
        Finds the article without using the DOI
        """
        fields = [k for k in article if (article[k])]

        possibles_year = self._year['None'].copy()
        if ('year' in fields) and (article['year'] in self._year):
            possibles_year = possibles_year.union(self._year[article['year']])
        if not possibles_year:
            return(None)

        possibles_authors = self._authors[1].intersection(possibles_year)
        nAuthors = len(article['authors'])
        if (nAuthors != 1) and (nAuthors in self._authors):
            possibles_authors = possibles_authors.union(
                self._authors[nAuthors])
        possibles = possibles_authors.intersection(possibles_year)
        if not possibles:
            return(None)

        if ('title' in fields):
            possibles_title = set()
            tokens = tokens_from_sentence(article['title'])
            for token in tokens:
                if token in self._title:
                    possibles_title = possibles_title.union(self._title[token])

            # if we didn't find anything, dont filter by it
            if possibles_title:
                possibles_title = possibles_title.union(self._title[0])
                possibles = possibles.intersection(possibles_title)
                if not possibles:
                    return(None)

        if ('authors' in fields):
            possibles_authorName = set()
            for author in article['authors']:
                for name in author.last_names:
                    token = name.lower()
                    if token in self._authorName:
                        possibles_authorName = possibles_authorName.union(
                            self._authorName[token])

            # if we didn't find anything, dont filter by it
            if possibles_authorName:
                possibles_authorName = possibles_authorName.union(
                    self._authorName[0])
                possibles = possibles.intersection(possibles_authorName)
                if not possibles:
                    return(None)

        if ('journal' in fields):
            possibles_journal = set()
            tokens = tokens_from_sentence(article['journal'])
            for token in tokens:
                if token in self._journal:
                    possibles_journal = possibles_journal.union(
                        self._journal[token])

            # if we didn't find anything, dont filter by it
            if possibles_journal:
                possibles_journal = possibles_journal.union(self._journal[0])
                possibles = possibles.intersection(possibles_journal)
                if not possibles:
                    return(None)

        for n in possibles:
            if same_article(self.node[n]['data'], article):
                return(n)
        return(None)

    def find(self, article: dict):
        """
        Finds the node corresponding to article a.
        Adds, in place, if it isn't in the graph.
        Replaces the node if the new article has a DOI.
        """
        # find by DOI
        if ('doi' in article) and (article['doi'] is not None):
            if (article['doi'] in self):
                return(article['doi'])
            # we have that DOI, but replaced with another. We follow the path!
            if (article['doi'] in self._equivalentDOIs):
                eq = article['doi']
                while eq not in self:
                    eq = self._equivalentDOIs[eq]
            # article might be there, just not with a DOI yet
            return(self.add(article, self._find_article_no_doi(article)))

        # article doesnt have a DOI
        n = self._find_article_no_doi(article)
        if (n is not None):
            return(n)
        else:
            return(self.add(article))
    # @profile

    def cocitation(self, count_label: str='count', copy_data: bool=True, min_cocitations: int=1)->nx.Graph:
        """
        Builds a co-citation network from the citation network.
        G[n1][n2][count_label] stores the co-citation count. 
        "copy_data" determins if the ['data'] structure from the references will
        also be copied to the co-citation network.
        min_cocitations limits the co-citation network to consider only pairs with >= min_cocitations.

        If the citation network has more than 500k articles, min_cocitations rises to 2.
        """
        G = nx.Graph()
        print('Building co-citation')
        useful_nodes = [n for n in self if list(self.successors(n))]
        useful_targets = [n for n in self if len(
            list(self.predecessors(n))) >= min_cocitations]
        for citing in tqdm(useful_nodes):
            cited = [x for x in self.successors(citing) if x in useful_targets]
            for w1, w2 in combinations(cited, 2):
                G.add_edge(w1, w2)
                if count_label not in G[w1][w2]:
                    G[w1][w2][count_label] = 0
                G[w1][w2][count_label] += 1

        # removing isolated nodes
        G.remove_nodes_from([n for n in G if len(list(G.neighbors(n))) == 0])

        if (copy_data):
            for n in G:
                G.node[n]['data'] = {**self.node[n]['data']}

        return(G)

    def compute_keywords(self, number_of_words=20, keyword_label: str='keywords', citing_keywords_label: str='citing-keywords'):
        """
        For each article that has an abstract, compute the corresponding keywords
        and add them to the 'data' dictionary, under "keyword_label".
        The whole dataset is necessary to compute the idf part of tf-idf.
        """
        corpus = []
        stop_words = list(set(stopwords.words('english')+['do', 'and', 'among', 'findings', 'is', 'in', 'results', 'an', 'as', 'are', 'only', 'number',
                                                          'have', 'using', 'research', 'find', 'from', 'for', 'to', 'with', 'than', 'since', 'most',
                                                          'also', 'which', 'between', 'has', 'more', 'be', 'we', 'that', 'but', 'it', 'how', 'approaches', 'approach',
                                                          'they', 'not', 'article', 'on', 'data', 'by', 'a', 'both', 'this', 'of', 'studies', 'lens', 'analysis', 'take', 'took',
                                                          'their', 'these', 'social', 'the', 'or', 'may', 'whether', 'them'', only', 'limit',
                                                          'implication', 'our', 'less', 'who', 'all', 'based', 'less', 'was', 'vital', 'taken', 'wide', 'view',
                                                          'its', 'new', 'one', 'use', 'these', 'focus', 'result', 'test', 'property', 'properties',
                                                          'finding', 'relationship', 'different', 'their', 'more', 'between', 'supposed', 'another',
                                                          'article', 'study', 'paper', 'research', 'sample', 'effect', 'case', 'argue', 'three', 'upon',
                                                          'affect', 'extent', 'when', 'implications', 'been', 'data', 'even', 'examine', 'toward', 'particularly',
                                                          'effects', 'analysis', 'into', 'support', 'show', 'within', 'what', 'were', 'per', 'focusing', 'focus',
                                                          'associated', 'suggest', 'those', 'over', 'however', 'while', 'indicate', 'about', 'second', 'first',
                                                          'terms', 'processes', 'tactics', 'strategies', 'de', 'involved', 'issues', 'successful', 'unfinished',
                                                          'such', 'other', 'because', 'can', 'both', 'n', 'find', 'using', 'have', 'not', 'role', 'rather',
                                                          'some', 'likely', 'findings', 'but', 'results', 'among', 'has', 'how', 'which', 'understand',
                                                          'they', 'be', 'i', 'two', 'than', 'how', 'which', 'be', 'across', 'also', 'it', 'through', 'at']))
        lemmatizer = WordNetLemmatizer()
        useful_nodes = [n for n in self if ('data' in self.node[n]) and (
            'abstract' in self.node[n]['data']) and (len(self.node[n]['data']['abstract']) > 0)]

        for n in useful_nodes:
            lemmas = [lemmatizer.lemmatize(word.translate(remove_punct), pos='s').lower(
            ) for word in word_tokenize(self.node[n]['data']['abstract'])]
            corpus.append(" ".join([word for word in lemmas if (
                word not in stop_words) and (word != '')]))

        cv = CountVectorizer(max_df=0.85, stop_words=stop_words,
                             max_features=10000, ngram_range=(1, 3))
        X = cv.fit_transform(corpus)
        tfidf_transformer = TfidfTransformer(use_idf=False)
        tfidf_transformer.fit(X)
        # get feature names
        feature_names = cv.get_feature_names()

        for i, doc in enumerate(corpus):
            # generate tf-idf for the given document
            tf_idf_vector = tfidf_transformer.transform(cv.transform([doc]))
            # sort the tf-idf vectors by descending order of scores
            sorted_items = sort_coo(tf_idf_vector.tocoo())
            # extract only the top n
            self.node[useful_nodes[i]]['data'][keyword_label] = extract_topn_from_vector(
                feature_names, sorted_items, number_of_words)

        print('Citing keywords')
        for n in tqdm(self):
            keywords = []
            used_documents = 0
            for citing in self.predecessors(n):
                # not all citing articles have abstracts
                if keyword_label in self.node[citing]['data']:
                    used_documents += 1
                    keywords.extend(self.node[citing]['data'][keyword_label])
            if keywords:
                keywords = _merge_keywords(keywords, used_documents)
                if len(keywords) > number_of_words:
                    keywords = sorted(keywords, key=lambda x: x[1], reverse=True)[
                        :number_of_words]
            self.node[n]['data'][citing_keywords_label] = keywords



from os.path import exists
from refcliq.util import cleanCurlyAround
import re
import networkx as nx
from fuzzywuzzy.process import extractOne
from fuzzywuzzy.fuzz import ratio
from time import sleep, monotonic
from tqdm import tqdm
import json
import googlemaps
import spacy
from titlecase import titlecase
from sys import stderr

CACHE = 'cache.json'

_US = ['AL', 'AK', 'AZ', 'AR', 'CA', 'CO', 'CT', 'DE', 'FL', 'GA', 'HI', 'ID', 'IL', 'IN', 'IA', 'KS', 'KY', 'LA', 'ME', 'MD', 'MA', 'MI', 'MN', 'MS', 'MO',
       'MT', 'NE', 'NV', 'NH', 'NJ', 'NM', 'NY', 'NC', 'ND', 'OH', 'OK', 'OR', 'PA', 'RI', 'SC', 'SD', 'TN', 'TX', 'UT', 'VT', 'VA', 'WA', 'WV', 'WI', 'WY']


def _computeFV(s: str)->list:
    """
        Computes a normalized histogram of the (common) letters in s.
    """
    trans = {}
    for i, c in enumerate('abcdefghiklmnopqrstuvwxyz'):
        trans[c] = i

    ret = [0, ]*len(trans)
    for c in s.lower():
        if (c in trans):
            ret[trans[c]] += 1
    return([x/sum(ret) for x in ret])


def _distanceFV(v1: list, v2: list)->float:
    """
        Computes the distance between two feature vectors
    """
    ret = 0
    for i in range(len(v1)):
        ret += abs(v1[i]-v2[i])
    return(ret/2.0)


def _find(d: dict, k: str, T: int=75)->str:
    """
        Finds the key most similar to k in d. 
        Returns none if no key is at least T similar (fuzzywuzzy ratio)
    """

    if k in d:
        return(k)
    if (k == ''):
        return(None)
    ret = extractOne(k, d.keys(), score_cutoff=T)
    if ret:
        return(ret[0])
    return(None)


def _remove_numbers(s: str)->str:
    """
        Removes all continuous sequences of characters that contain numbers. 
        Ex. 11215, F-93160, etc...
    """
    return(','.join([' '.join([word for word in x.split() if not any([c.isdigit() for c in word])]) for x in s.split(',')]).replace(',,', ',').replace(',', ', '))


def _count_useful_parts(address_parts: list, google_result: dict, T: int=75)->int:
    """
        Determines how many address_parts are needed for city-level geocoding.
    """
    ret = 0
    for a in address_parts:
        found = False
        for p in google_result["address_components"]:
            if (ratio(a.lower(), p['long_name'].lower()) > T) or (ratio(a.lower(), p['short_name'].lower()) > T):
                found = True
                break
        if found:
            ret += 1

    return(ret)


def _filter_geocode(google_result: dict):
    useful = set(['postal_town', 'administrative_area_level_1',
                  'administrative_area_level_2', 'country', 'locality'])

    to_remove = []
    for i, part in enumerate(google_result["address_components"]):
        if len(set(part['types']).intersection(useful)) == 0:
            to_remove.append(i)

    for i in sorted(to_remove, reverse=True):
        del(google_result["address_components"][i])

    return(google_result)


class ArticleGeoCoder:
    def __init__(self, google_key: str=''):
        self._cache = {}
        if (google_key != ''):
            self._gmaps = googlemaps.Client(key=google_key)
        else:
            self._gmaps = None

        # those always pose a problem
        self._parts_by_country = {'peoples r china': 3, 'United States': 3,
                                  'czech republic': 2, 'ireland': 2, 'norway': 2, 'italy': 2, 'finland': 2}
        self._last_request = monotonic()  # keeps track of the last time we used nominatim
        self._outgoing_calls = 0

        if exists(CACHE):
            with open(CACHE, 'r') as fin:
                data = json.load(fin)
                self._cache = data['cache']
                self._parts_by_country.update(data['parts'])

    def _save_state(self):
        # 'fv':self._cacheFV
        to_save = {'cache': self._cache, 'parts': self._parts_by_country}
        with open('cache.json', 'w') as fout:
            json.dump(to_save, fout, indent=4, sort_keys=True)

    def add_authors_location_inplace(self, G: nx.Graph):
        """
            For every node of G (a reference in the network), finds the
            coordinates based from the 'Affiliation' bibtex field, if present.
            _Alters the data of G_.
        """
        #https://stackoverflow.com/questions/53383352/spacy-and-spacy-models-in-setup-py
        try:
            nlp = spacy.load('en_core_web_sm')
        except OSError:
            print('Downloading language model for spaCy\n'
                "(don't worry, this will only happen once)", file=stderr)
            from spacy.cli import download
            download('en_core_web_sm')
            nlp = spacy.load('en_core_web_sm')
        
        trees = {}
        print('Compiling addresses')
        for n in tqdm(G):
            G.node[n]['data']['countries'] = []
            G.node[n]['data']['coordinates'] = []
            addresses = []
            if ('data' in G.node[n]) and ('Affiliation' in G.node[n]['data']) and (G.node[n]['data']['Affiliation'] is not None) and (len(G.node[n]['data']['Affiliation']) > 0):
                aff = G.node[n]['data']['Affiliation'].replace('(Reprint Author)', '').replace(
                    ".,", ',').replace("'", '')  # O'lastname / Jesusm. / Redone.  mess sentence identification
                doc = nlp(aff)
                add = ''
                for sent in doc.sents:
                    if sent.text.endswith('.'):
                        current = _remove_numbers(
                            (add+' '+sent.text).strip().replace('  ', ' '))
                        if current.endswith('.'):
                            current = current[:-1]
                        # special cases: NY 10012 USA / LOS ANGELES,CA.
                        vals = [x.strip() for x in current.split(',')]
                        # ,CA.
                        if (len(vals[-1]) == 2) and vals[-1].upper() in _US:
                            addresses.append(
                                [titlecase(vals[-2]), vals[-1], 'United States'])
                        elif (vals[-1].endswith(' USA')):
                            addresses.append(
                                [titlecase(vals[-2]), vals[-1].split(' ')[0], 'United States'])
                        else:
                            if len(vals) > 3:  # name, dept, etc
                                addresses.append([titlecase(x)
                                                  for x in vals[-3:]])
                            else:
                                addresses.append([titlecase(x) for x in vals])
                        add = ''
                    else:
                        add = add + ' ' + sent.text

                for vals in addresses:
                    G.node[n]['data']['countries'].append(vals[-1])
                    if self._gmaps is not None:
                        v = [x.lower() for x in vals]
                        if len(v) < 3:
                            v = ['', ]*(3-len(v)) + v
                        # not really city / state / country, but it is easier to
                        # code with names
                        # there is probably a shorter way of doing this using defaultdicts
                        country = _find(trees, v[-1])
                        if country is None:
                            country = v[-1]
                            trees[country] = {}

                        state = _find(trees[country], v[-2])
                        if state is None:
                            state = v[-2]
                            trees[country][state] = {}

                        city = _find(trees[country][state], v[-3])
                        if city is None:
                            city = v[-3]
                            trees[country][state][city] = []

                        trees[country][state][city].append(n)

        
        for country in trees:
            cached = _find(self._parts_by_country, country)

            if cached is None:
                i = 0
                j = 0
                geo = []
                while not geo:
                    try:
                        sample_state = list(trees[country].keys())[i]
                        sample_city = list(trees[country][sample_state])[j]
                        parts = [sample_city, sample_state, country]
                        geo = self._google(parts)
                        j += 1
                        if j == len(trees[country][sample_state].keys()):
                            j = 0
                            i += 1
                            if i == len(trees[country]):
                                # didn't find anything, run all... very unlikely. Very.
                                to_use = 3
                    except:
                        print('Could not find anything using "{0}" as a country, please check the affiliation field.'.format(
                            country))
                        print(
                            '(This can happen when a author name has a dot at the end of a non-abreviation.)')
                        # print(trees[country])
                        raise
                if geo:
                    self._parts_by_country[country] = _count_useful_parts(
                        parts, geo)
                    self._save_state()

        print('Getting coordinates')
        for country in tqdm(trees):
            cached = _find(self._parts_by_country, country)
            to_use = self._parts_by_country[cached]

            for state in trees[country]:
                for city in trees[country][state]:
                    parts = [city, state, country][-to_use:]
                    parts = ['', ]*(3-len(parts)) + parts
                    geo = self._cache_search(parts)
                    if geo is None:
                        geo = self._google(parts)
                        if not geo:
                            continue
                        self._cache_add(parts, geo)

                    for n in trees[country][state][city]:
                        G.node[n]['data']['coordinates'].append(
                            [geo['geometry']['location']['lng'], geo['geometry']['location']['lat']])

        return(G)

    def _cache_search(self, address_parts: list):
        vals = ['', ]*(3-(len(address_parts))) + address_parts

        country = _find(self._cache, vals[-1])
        if country is None:
            return(None)

        state = _find(self._cache[country], vals[-2])
        if state is None:
            return(None)

        city = _find(self._cache[country][state], vals[-3])
        if city is None:
            return(None)

        return(self._cache[country][state][city])

    def _cache_add(self, address_parts: list, google_geocode: dict):
        vals = ['', ]*(3-(len(address_parts))) + address_parts

        country = _find(self._cache, vals[-1])
        if country is None:
            country = vals[-1]
            self._cache[country] = {}

        state = _find(self._cache[country], vals[-2])
        if state is None:
            state = vals[-2]
            self._cache[country][state] = {}

        city = _find(self._cache[country][state], vals[-3])
        if city is None:
            city = vals[-3]

        self._cache[country][state][city] = google_geocode
        self._save_state()

    def _google(self, address_parts: list)->list:
        """
            Queries google's geocoding service for the address.
            Limited to 50 queries per second. A key is necessary.
            Returns (lng,lat) for the point.
        """
        minTime = 1/50
        delta = monotonic() - self._last_request
        res = None
        address = ', '.join([x for x in address_parts if x != ''])
        if delta <= minTime:
            sleep(minTime-delta)

        res = self._gmaps.geocode(address)
        # with open('testing.json','r') as fin:
        #     res = json.load(fin)
        if res:
            res = _filter_geocode(res[0])

        # print('google ', address)
        self._outgoing_calls += 1
        self._last_request = monotonic()
        return(res)



import setuptools
from glob import glob
from os.path import basename, dirname, join, splitext
from os import walk


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="refcliq",
    version="0.1.1",
    author="Fabio Dias",
    author_email="fabio.dias@gmail.com",
    description="Community analysis in bibliographical references",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fabioasdias/RefCliq",
    packages=['refcliq'],
    package_dir={'refcliq': 'src/refcliq'},
    package_data={'refcliq': ['template/*',
                              'template/static/*',
                              'template/static/css/*',
                              'template/static/js/*']},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    # include_package_data = True,    
    scripts = ['rc_cluster.py', 'rc_vis.py'],
    install_requires = [
        "python-louvain>=0.13",
        "numpy>=1.16.2",
        "pybtex>=0.22.2",
        "nltk>=3.4",
        "tqdm>=4.31.1",
        "titlecase>=0.12.0",
        "fuzzywuzzy[speedup]>=0.17.0",
        "klepto>=0.1.6",
        "h5py>=2.9.0",
        "spacy>=2.1.3",
        # "en_core_web_sm @ https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-2.1.0/en_core_web_sm-2.1.0.tar.gz",
        "googlemaps>=3.0.2",
        "scikit-learn>=0.20.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)